#include <graph/server/gui/chart/chart.h>
#include <graph/server/gui/mainwindow.h>
#include <qbrush.h>
#include <qnamespace.h>
#include <qpoint.h>
#include <qpolygon.h>
#include <qsize.h>
#include <QPainter>
#include <iostream>
#include <iterator>
#include <limits>
#include <string>
#include "graph/plugins/plugin.h"

#define MIN(a, b) (a < b ? a : b)
#define MAX(a, b) (a > b ? a : b)

namespace Graph::GUI {
ChartWidget::ChartWidget(QWidget* parent, const chart_spec_t& spec) : QWidget(parent), spec(spec) {
  setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

  setMinimumSize(QSize(600, 350));

  // context menu
  setContextMenuPolicy(Qt::ContextMenuPolicy::CustomContextMenu);
  connect(this, &QWidget::customContextMenuRequested, this, [this](const QPoint& pos) {
    ContextMenu menu(nullptr, this);
    menu.exec(mapToGlobal(pos));
  });
}

ChartWidget::~ChartWidget() {}

void ChartWidget::addData(const char* seriesName, const point_t* points, unsigned long count) {
  QMutexLocker locker(&mutex);

  double minX = std::numeric_limits<double>::max();
  double minY = std::numeric_limits<double>::max();
  double maxX = std::numeric_limits<double>::min();
  double maxY = std::numeric_limits<double>::min();

  if (!spec.keep_all) {
    if (count >= spec.keep_limit) {
      this->points.clear();
    } else if (this->points.size() + count > spec.keep_limit) {
      std::list<point_t>::iterator end = std::next(this->points.begin(), this->points.size() - (spec.keep_limit - count));
      this->points.erase(this->points.begin(), end);
    }
  }

  for (const point_t& point : this->points) {
    if (spec.axis.x.auto_limit) {
      minX = MIN(minX, point.x);
      maxX = MAX(maxX, point.x);
    }
    if (spec.axis.y.auto_limit) {
      minY = MIN(minY, point.y);
      maxY = MAX(maxY, point.y);
    }
  }

  for (unsigned long i = spec.keep_all ? 0 : (count < spec.keep_limit ? 0 : count - spec.keep_limit); i < count; i++) {
    if (decimateState >= spec.decimate)
      decimateState = 0;
    if (decimateState++ % spec.decimate != 0)
      continue;

    this->points.push_back(points[i]);

    if (spec.axis.x.auto_limit) {
      minX = MIN(minX, points[i].x);
      maxX = MAX(maxX, points[i].x);
    }
    if (spec.axis.y.auto_limit) {
      minY = MIN(minY, points[i].y);
      maxY = MAX(maxY, points[i].y);
    }
  }

  if (spec.axis.x.auto_limit) {
    spec.axis.x.min = minX;
    spec.axis.x.max = maxX;
  }
  if (spec.axis.y.auto_limit) {
    spec.axis.y.min = minY;
    spec.axis.y.max = maxY;
  }

  update();
}

void ChartWidget::clear() {
  QMutexLocker locker(&mutex);
  points.clear();
}

QSize ChartWidget::sizeHint() const {
  if (width() < minimumWidth() || height() < minimumHeight())
    return minimumSize();
  return QSize(width(), height());
}

QPointF ChartWidget::pointToCoord(const point_t& point) {
  return QPointF((point.x - spec.axis.x.min) * scaleX + marginX, height() - (point.y - spec.axis.y.min) * scaleY - marginY);
}

void ChartWidget::paintEvent(QPaintEvent* e) {
  QMutexLocker locker(&mutex);

  QPainter painter(this);

  // background
  painter.fillRect(QRectF(0, 0, width(), height()), Qt::lightGray);
  painter.fillRect(QRectF(marginX, marginY, width() - 2 * marginX, height() - 2 * marginY), Qt::white);

  // title
  painter.setPen(Qt::black);
  painter.drawText(QRect(0, 0, width(), marginY), Qt::AlignCenter, spec.title);

  // limits
  // y max-mid-min
  painter.drawText(QRect(0, marginY, marginX, marginY), Qt::AlignCenter, QString::number(spec.axis.y.max));
  painter.drawText(QRect(0, height() / 2 - marginY / 2, marginX, marginY), Qt::AlignCenter, QString::number((spec.axis.y.max + spec.axis.y.min) / 2));
  painter.drawText(QRect(0, height() - marginY * 2, marginX, marginY), Qt::AlignCenter, QString::number(spec.axis.y.min));

  // x min-mid-max
  painter.drawText(QRect(marginX, height() - marginY, marginX, marginY), Qt::AlignCenter, QString::number(spec.axis.x.min));
  painter.drawText(QRect(width() / 2 - marginX / 2, height() - marginY, marginX, marginY), Qt::AlignCenter,
                   QString::number((spec.axis.x.min + spec.axis.x.max) / 2));
  painter.drawText(QRect(width() - marginX * 2, height() - marginY, marginX, marginY), Qt::AlignCenter, QString::number(spec.axis.x.max));

  // grid
  if (spec.grid) {
    painter.setPen(Qt::lightGray);

    int innerHeight = height() - 2 * marginY;
    painter.drawLine(marginX, marginY + innerHeight / 4, width() - marginX, marginY + innerHeight / 4);
    painter.drawLine(marginX, marginY + innerHeight / 2, width() - marginX, marginY + innerHeight / 2);
    painter.drawLine(marginX, marginY + 3 * innerHeight / 4, width() - marginX, marginY + 3 * innerHeight / 4);

    int innerWidth = width() - 2 * marginX;
    painter.drawLine(marginX + innerWidth / 4, marginY, marginX + innerWidth / 4, height() - marginY);
    painter.drawLine(marginX + innerWidth / 2, marginY, marginX + innerWidth / 2, height() - marginY);
    painter.drawLine(marginX + 3 * innerWidth / 4, marginY, marginX + 3 * innerWidth / 4, height() - marginY);
  }

  // points
  scaleX = (width() - marginX * 2) / (spec.axis.x.max - spec.axis.x.min);
  scaleY = (height() - marginY * 2) / (spec.axis.y.max - spec.axis.y.min);

  painter.setPen(QPen(Qt::black, 1));

  QPolygonF polygon;
  for (const point_t point : points) {
    // TODO: only paint visible points
    //       (+ 1-1 before and after)
    polygon << pointToCoord(point);
  }

  painter.drawPolyline(polygon);
}

ContextMenu::ContextMenu(QWidget* parent, ChartWidget* chartWidget) : QMenu(parent) {
  remove = new QAction(tr("&Remove"), this);
  connect(remove, &QAction::triggered, this, [chartWidget]() { MainWindow::getInstance()->removeChart(chartWidget); });
  addAction(remove);

  clear = new QAction(tr("&Clear"), this);
  connect(clear, &QAction::triggered, this, [chartWidget]() { chartWidget->clear(); });
  addAction(clear);

  edit = new QAction(tr("&Edit"), this);
  connect(edit, &QAction::triggered, this, []() {
    // TODO: implement chart editing
  });
  addAction(edit);
}

ContextMenu::~ContextMenu() {}
} // namespace Graph::GUI